# -*- coding: utf-8; -*-
"""grow datasync_change.source, consumer

Revision ID: 8b02c6eaf318
Revises: 830a20917c49
Create Date: 2018-07-10 18:55:30.539083

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '8b02c6eaf318'
down_revision = u'830a20917c49'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # datasync_change
    op.alter_column('datasync_change', 'source', type_=sa.String(length=50))
    op.alter_column('datasync_change', 'consumer', type_=sa.String(length=50))


def downgrade():

    # datasync_change
    op.alter_column('datasync_change', 'consumer', type_=sa.String(length=20))
    op.alter_column('datasync_change', 'source', type_=sa.String(length=20))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CSV Reports
"""

from __future__ import unicode_literals, absolute_import

import csv

import six

from rattail.reporting import Report
from rattail.csvutil import UnicodeDictWriter


class CSVReport(Report):
    """
    Generic report which knows how to write CSV output file.

    .. attr:: output_fields

       Simple list of field names which should be included in the output file.

    """
    output_fields = []

    def make_filename(self, session, params, data, **kwargs):
        return "{}.csv".format(self.name)

    def get_output_fields(self, params):
        return list(self.output_fields)

    def write_file(self, session, params, data, path, progress=None, **kwargs):
        """
        Write a basic CSV output file with the given data.  Requires
        :attr:`output_fields` attribute to be set to work correctly.
        """
        fields = self.get_output_fields(params)

        if six.PY2:
            csv_file = open(path, 'wb')
            writer = UnicodeDictWriter(csv_file, fields, encoding='utf_8')
        else: # PY3
            csv_file = open(path, 'wt', encoding='utf_8')
            writer = csv.DictWriter(csv_file, fields)

        self.write_file_header(writer)

        def writerow(row, i):
            writer.writerow(row)

        self.progress_loop(writerow, data, progress,
                           message="Generating CSV output")
        csv_file.close()

    def write_file_header(self, writer, **kwargs):
        writer.writeheader()



# -*- coding: utf-8; -*-

from __future__ import unicode_literals, absolute_import

import os
import shutil
from unittest import TestCase

from rattail import labels as mod
from rattail.config import make_config
from rattail.exceptions import LabelPrintingError


class TestLabelPrinter(TestCase):

    def setUp(self):
        self.config = self.make_config()
        self.printer = self.make_printer()

    def make_config(self):
        return make_config([], extend=False)

    def make_printer(self):
        return mod.LabelPrinter(self.config)

    def test_print_labels(self):
        labels = []

        # not implemented by default
        self.assertRaises(NotImplementedError, self.printer.print_labels, labels)


class TestCommandFilePrinter(TestCase):

    def setUp(self):
        self.config = self.make_config()
        self.printer = self.make_printer()

    def make_config(self):
        config = make_config([], extend=False)
        config.setdefault('rattail', 'timezone.default', 'America/Chicago')
        return config

    def make_printer(self):
        printer = mod.CommandFilePrinter(self.config)
        printer.formatter = mod.CommandFormatter(self.config, template="")
        return printer

    def test_print_labels(self):
        app = self.config.get_app()
        labels = []

        # output_dir is required setting
        self.assertRaises(LabelPrintingError, self.printer.print_labels, labels)

        # okay now with output_dir
        outdir = app.make_temp_dir()
        self.printer.output_dir = outdir
        path = self.printer.print_labels(labels)
        self.assertEqual(os.path.dirname(path), outdir)

        # also can override output_dir by passing it in call
        outdir2 = app.make_temp_dir()
        path2 = self.printer.print_labels(labels, output_dir=outdir2)
        self.assertEqual(os.path.dirname(path2), outdir2)

        shutil.rmtree(outdir)
        shutil.rmtree(outdir2)



# -*- coding: utf-8; -*-
"""rename batch_purchase

Revision ID: 82d3c2d9830d
Revises: 6b06ebea979c
Create Date: 2021-02-02 15:34:14.574117

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '82d3c2d9830d'
down_revision = '6b06ebea979c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # purchase
    op.drop_constraint('purchase_purchase_fk_buyer', 'purchase', type_='foreignkey')
    op.create_foreign_key('purchase_fk_buyer',
                          'purchase', 'employee',
                          ['buyer_uuid'], ['uuid'])

    # remove references to `purchase_batch`
    op.drop_constraint('purchase_batch_row_fk_batch_uuid', 'purchase_batch_row', type_='foreignkey')

    # batch_purchase
    op.drop_constraint('purchase_batch_fk_created_by', 'purchase_batch', type_='foreignkey')
    op.drop_constraint('purchase_batch_fk_cognized_by', 'purchase_batch', type_='foreignkey')
    op.drop_constraint('purchase_batch_fk_executed_by', 'purchase_batch', type_='foreignkey')
    op.drop_constraint('purchase_batch_fk_store', 'purchase_batch', type_='foreignkey')
    op.drop_constraint('purchase_batch_fk_vendor', 'purchase_batch', type_='foreignkey')
    op.drop_constraint('purchase_batch_fk_department', 'purchase_batch', type_='foreignkey')
    op.drop_constraint('purchase_batch_purchase_fk_buyer', 'purchase_batch', type_='foreignkey')
    op.drop_constraint('purchase_batch_fk_purchase', 'purchase_batch', type_='foreignkey')
    op.drop_constraint('purchase_batch_fk_truck_dump_batch', 'purchase_batch', type_='foreignkey')
    op.rename_table('purchase_batch', 'batch_purchase')
    op.create_foreign_key('batch_purchase_fk_created_by',
                          'batch_purchase', 'user',
                          ['created_by_uuid'], ['uuid'])
    op.create_foreign_key('batch_purchase_fk_cognized_by',
                          'batch_purchase', 'user',
                          ['cognized_by_uuid'], ['uuid'])
    op.create_foreign_key('batch_purchase_fk_executed_by',
                          'batch_purchase', 'user',
                          ['executed_by_uuid'], ['uuid'])
    op.create_foreign_key('batch_purchase_fk_store',
                          'batch_purchase', 'store',
                          ['store_uuid'], ['uuid'])
    op.create_foreign_key('batch_purchase_fk_vendor',
                          'batch_purchase', 'vendor',
                          ['vendor_uuid'], ['uuid'])
    op.create_foreign_key('batch_purchase_fk_department',
                          'batch_purchase', 'department',
                          ['department_uuid'], ['uuid'])
    op.create_foreign_key('batch_purchase_fk_buyer',
                          'batch_purchase', 'employee',
                          ['buyer_uuid'], ['uuid'])
    op.create_foreign_key('batch_purchase_fk_purchase',
                          'batch_purchase', 'purchase',
                          ['purchase_uuid'], ['uuid'])
    op.create_foreign_key('batch_purchase_fk_truck_dump_batch',
                          'batch_purchase', 'batch_purchase',
                          ['truck_dump_batch_uuid'], ['uuid'])

    # batch_purchase_row
    op.drop_constraint('purchase_batch_row_fk_product', 'purchase_batch_row', type_='foreignkey')
    op.drop_constraint('purchase_batch_row_fk_item', 'purchase_batch_row', type_='foreignkey')
    op.rename_table('purchase_batch_row', 'batch_purchase_row')
    op.create_foreign_key('batch_purchase_row_fk_batch_uuid',
                          'batch_purchase_row', 'batch_purchase',
                          ['batch_uuid'], ['uuid'])
    op.create_foreign_key('batch_purchase_row_fk_product',
                          'batch_purchase_row', 'product',
                          ['product_uuid'], ['uuid'])
    op.create_foreign_key('batch_purchase_row_fk_item',
                          'batch_purchase_row', 'purchase_item',
                          ['item_uuid'], ['uuid'])

    # batch_purchase_row_claim
    op.drop_constraint('purchase_batch_row_claim_fk_claiming_row', 'purchase_batch_row_claim', type_='foreignkey')
    op.drop_constraint('purchase_batch_row_claim_fk_claimed_row', 'purchase_batch_row_claim', type_='foreignkey')
    op.rename_table('purchase_batch_row_claim', 'batch_purchase_row_claim')
    op.create_foreign_key('batch_purchase_row_claim_fk_claiming_row',
                          'batch_purchase_row_claim', 'batch_purchase_row',
                          ['claiming_row_uuid'], ['uuid'])
    op.create_foreign_key('batch_purchase_row_claim_fk_claimed_row',
                          'batch_purchase_row_claim', 'batch_purchase_row',
                          ['claimed_row_uuid'], ['uuid'])

    # batch_purchase_credit
    op.drop_constraint('purchase_batch_credit_fk_store', 'purchase_batch_credit', type_='foreignkey')
    op.drop_constraint('purchase_batch_credit_fk_vendor', 'purchase_batch_credit', type_='foreignkey')
    op.drop_constraint('purchase_batch_credit_fk_product', 'purchase_batch_credit', type_='foreignkey')
    op.drop_constraint('purchase_batch_credit_fk_mispick_product', 'purchase_batch_credit', type_='foreignkey')
    op.drop_constraint('purchase_batch_credit_fk_row', 'purchase_batch_credit', type_='foreignkey')
    op.rename_table('purchase_batch_credit', 'batch_purchase_credit')
    op.create_foreign_key('batch_purchase_credit_fk_store',
                          'batch_purchase_credit', 'store',
                          ['store_uuid'], ['uuid'])
    op.create_foreign_key('batch_purchase_credit_fk_vendor',
                          'batch_purchase_credit', 'vendor',
                          ['vendor_uuid'], ['uuid'])
    op.create_foreign_key('batch_purchase_credit_fk_product',
                          'batch_purchase_credit', 'product',
                          ['product_uuid'], ['uuid'])
    op.create_foreign_key('batch_purchase_credit_fk_mispick_product',
                          'batch_purchase_credit', 'product',
                          ['mispick_product_uuid'], ['uuid'])
    op.create_foreign_key('batch_purchase_credit_fk_row',
                          'batch_purchase_credit', 'batch_purchase_row',
                          ['row_uuid'], ['uuid'])


def downgrade():

    # purchase
    op.drop_constraint('purchase_fk_buyer', 'purchase', type_='foreignkey')
    op.create_foreign_key('purchase_purchase_fk_buyer',
                          'purchase', 'employee',
                          ['buyer_uuid'], ['uuid'])

    # remove references to `batch_purchase`
    op.drop_constraint('batch_purchase_row_fk_batch_uuid', 'batch_purchase_row', type_='foreignkey')

    # batch_purchase
    op.drop_constraint('batch_purchase_fk_created_by', 'batch_purchase', type_='foreignkey')
    op.drop_constraint('batch_purchase_fk_cognized_by', 'batch_purchase', type_='foreignkey')
    op.drop_constraint('batch_purchase_fk_executed_by', 'batch_purchase', type_='foreignkey')
    op.drop_constraint('batch_purchase_fk_store', 'batch_purchase', type_='foreignkey')
    op.drop_constraint('batch_purchase_fk_vendor', 'batch_purchase', type_='foreignkey')
    op.drop_constraint('batch_purchase_fk_department', 'batch_purchase', type_='foreignkey')
    op.drop_constraint('batch_purchase_fk_buyer', 'batch_purchase', type_='foreignkey')
    op.drop_constraint('batch_purchase_fk_purchase', 'batch_purchase', type_='foreignkey')
    op.drop_constraint('batch_purchase_fk_truck_dump_batch', 'batch_purchase', type_='foreignkey')
    op.rename_table('batch_purchase', 'purchase_batch')
    op.create_foreign_key('purchase_batch_fk_created_by',
                          'purchase_batch', 'user',
                          ['created_by_uuid'], ['uuid'])
    op.create_foreign_key('purchase_batch_fk_cognized_by',
                          'purchase_batch', 'user',
                          ['cognized_by_uuid'], ['uuid'])
    op.create_foreign_key('purchase_batch_fk_executed_by',
                          'purchase_batch', 'user',
                          ['executed_by_uuid'], ['uuid'])
    op.create_foreign_key('purchase_batch_fk_store',
                          'purchase_batch', 'store',
                          ['store_uuid'], ['uuid'])
    op.create_foreign_key('purchase_batch_fk_vendor',
                          'purchase_batch', 'vendor',
                          ['vendor_uuid'], ['uuid'])
    op.create_foreign_key('purchase_batch_fk_department',
                          'purchase_batch', 'department',
                          ['department_uuid'], ['uuid'])
    op.create_foreign_key('purchase_batch_purchase_fk_buyer',
                          'purchase_batch', 'employee',
                          ['buyer_uuid'], ['uuid'])
    op.create_foreign_key('purchase_batch_fk_purchase',
                          'purchase_batch', 'purchase',
                          ['purchase_uuid'], ['uuid'])
    op.create_foreign_key('purchase_batch_fk_truck_dump_batch',
                          'purchase_batch', 'purchase_batch',
                          ['truck_dump_batch_uuid'], ['uuid'])

    # batch_purchase_row
    op.drop_constraint('batch_purchase_row_fk_product', 'batch_purchase_row', type_='foreignkey')
    op.drop_constraint('batch_purchase_row_fk_item', 'batch_purchase_row', type_='foreignkey')
    op.rename_table('batch_purchase_row', 'purchase_batch_row')
    op.create_foreign_key('purchase_batch_row_fk_batch_uuid',
                          'purchase_batch_row', 'purchase_batch',
                          ['batch_uuid'], ['uuid'])
    op.create_foreign_key('purchase_batch_row_fk_product',
                          'purchase_batch_row', 'product',
                          ['product_uuid'], ['uuid'])
    op.create_foreign_key('purchase_batch_row_fk_item',
                          'purchase_batch_row', 'purchase_item',
                          ['item_uuid'], ['uuid'])

    # batch_purchase_row_claim
    op.drop_constraint('batch_purchase_row_claim_fk_claiming_row', 'batch_purchase_row_claim', type_='foreignkey')
    op.drop_constraint('batch_purchase_row_claim_fk_claimed_row', 'batch_purchase_row_claim', type_='foreignkey')
    op.rename_table('batch_purchase_row_claim', 'purchase_batch_row_claim')
    op.create_foreign_key('purchase_batch_row_claim_fk_claiming_row',
                          'purchase_batch_row_claim', 'purchase_batch_row',
                          ['claiming_row_uuid'], ['uuid'])
    op.create_foreign_key('purchase_batch_row_claim_fk_claimed_row',
                          'purchase_batch_row_claim', 'purchase_batch_row',
                          ['claimed_row_uuid'], ['uuid'])

    # batch_purchase_credit
    op.drop_constraint('batch_purchase_credit_fk_store', 'batch_purchase_credit', type_='foreignkey')
    op.drop_constraint('batch_purchase_credit_fk_vendor', 'batch_purchase_credit', type_='foreignkey')
    op.drop_constraint('batch_purchase_credit_fk_product', 'batch_purchase_credit', type_='foreignkey')
    op.drop_constraint('batch_purchase_credit_fk_mispick_product', 'batch_purchase_credit', type_='foreignkey')
    op.drop_constraint('batch_purchase_credit_fk_row', 'batch_purchase_credit', type_='foreignkey')
    op.rename_table('batch_purchase_credit', 'purchase_batch_credit')
    op.create_foreign_key('purchase_batch_credit_fk_store',
                          'purchase_batch_credit', 'store',
                          ['store_uuid'], ['uuid'])
    op.create_foreign_key('purchase_batch_credit_fk_vendor',
                          'purchase_batch_credit', 'vendor',
                          ['vendor_uuid'], ['uuid'])
    op.create_foreign_key('purchase_batch_credit_fk_product',
                          'purchase_batch_credit', 'product',
                          ['product_uuid'], ['uuid'])
    op.create_foreign_key('purchase_batch_credit_fk_mispick_product',
                          'purchase_batch_credit', 'product',
                          ['mispick_product_uuid'], ['uuid'])
    op.create_foreign_key('purchase_batch_credit_fk_row',
                          'purchase_batch_credit', 'purchase_batch_row',
                          ['row_uuid'], ['uuid'])



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Settings API
"""

import warnings

from rattail.db import Session, model


def get_setting(session, name):
    """
    Returns a setting value from the database.
    """
    warnings.warn("function is deprecated, please use "
                  "AppHandler.get_setting() instead",
                  DeprecationWarning, stacklevel=2)

    local_session = False
    if session is None:
        session = Session()
        local_session = True
    setting = session.get(model.Setting, name)
    value = None if setting is None else setting.value
    if local_session:
        session.close()
    return value


def save_setting(session, name, value):
    """
    Saves a setting to the database.
    """
    warnings.warn("function is deprecated, please use "
                  "AppHandler.save_setting() instead",
                  DeprecationWarning, stacklevel=2)

    local_session = False
    if session is None:
        session = Session()
        local_session = True
    setting = session.get(model.Setting, name)
    if not setting:
        setting = model.Setting(name=name)
        session.add(setting)
    setting.value = value
    if local_session:
        session.commit()
        session.close()



# -*- coding: utf-8; -*-
"""rename batch_vendorcatalog

Revision ID: 89e2406859be
Revises: 6b5a277cc31b
Create Date: 2021-02-02 12:44:00.529051

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '89e2406859be'
down_revision = '6b5a277cc31b'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # remove references to `vendor_catalog`
    op.drop_constraint('vendor_catalog_row_fk_batch_uuid', 'vendor_catalog_row', type_='foreignkey')

    # batch_vendorcatalog
    op.drop_constraint('vendor_catalog_fk_created_by', 'vendor_catalog', type_='foreignkey')
    op.drop_constraint('vendor_catalog_fk_cognized_by', 'vendor_catalog', type_='foreignkey')
    op.drop_constraint('vendor_catalog_fk_executed_by', 'vendor_catalog', type_='foreignkey')
    op.drop_constraint('vendor_catalog_fk_vendor', 'vendor_catalog', type_='foreignkey')
    op.rename_table('vendor_catalog', 'batch_vendorcatalog')
    op.create_foreign_key('batch_vendorcatalog_fk_created_by',
                          'batch_vendorcatalog', 'user',
                          ['created_by_uuid'], ['uuid'])
    op.create_foreign_key('batch_vendorcatalog_fk_cognized_by',
                          'batch_vendorcatalog', 'user',
                          ['cognized_by_uuid'], ['uuid'])
    op.create_foreign_key('batch_vendorcatalog_fk_executed_by',
                          'batch_vendorcatalog', 'user',
                          ['executed_by_uuid'], ['uuid'])
    op.create_foreign_key('batch_vendorcatalog_fk_vendor',
                          'batch_vendorcatalog', 'vendor',
                          ['vendor_uuid'], ['uuid'])

    # batch_vendorcatalog_row
    op.drop_constraint('vendor_catalog_row_fk_product', 'vendor_catalog_row', type_='foreignkey')
    op.drop_constraint('vendor_catalog_row_fk_cost', 'vendor_catalog_row', type_='foreignkey')
    op.rename_table('vendor_catalog_row', 'batch_vendorcatalog_row')
    op.create_foreign_key('batch_vendorcatalog_row_fk_batch_uuid',
                          'batch_vendorcatalog_row', 'batch_vendorcatalog',
                          ['batch_uuid'], ['uuid'])
    op.create_foreign_key('batch_vendorcatalog_row_fk_product',
                          'batch_vendorcatalog_row', 'product',
                          ['product_uuid'], ['uuid'])
    op.create_foreign_key('batch_vendorcatalog_row_fk_cost',
                          'batch_vendorcatalog_row', 'product_cost',
                          ['cost_uuid'], ['uuid'])


def downgrade():

    # remove references to `batch_vendorcatalog`
    op.drop_constraint('batch_vendorcatalog_row_fk_batch_uuid', 'batch_vendorcatalog_row', type_='foreignkey')

    # batch_vendorcatalog
    op.drop_constraint('batch_vendorcatalog_fk_created_by', 'batch_vendorcatalog', type_='foreignkey')
    op.drop_constraint('batch_vendorcatalog_fk_cognized_by', 'batch_vendorcatalog', type_='foreignkey')
    op.drop_constraint('batch_vendorcatalog_fk_executed_by', 'batch_vendorcatalog', type_='foreignkey')
    op.drop_constraint('batch_vendorcatalog_fk_vendor', 'batch_vendorcatalog', type_='foreignkey')
    op.rename_table('batch_vendorcatalog', 'vendor_catalog')
    op.create_foreign_key('vendor_catalog_fk_created_by',
                          'vendor_catalog', 'user',
                          ['created_by_uuid'], ['uuid'])
    op.create_foreign_key('vendor_catalog_fk_cognized_by',
                          'vendor_catalog', 'user',
                          ['cognized_by_uuid'], ['uuid'])
    op.create_foreign_key('vendor_catalog_fk_executed_by',
                          'vendor_catalog', 'user',
                          ['executed_by_uuid'], ['uuid'])
    op.create_foreign_key('vendor_catalog_fk_vendor',
                          'vendor_catalog', 'vendor',
                          ['vendor_uuid'], ['uuid'])

    # batch_vendorcatalog_row
    op.drop_constraint('batch_vendorcatalog_row_fk_product', 'batch_vendorcatalog_row', type_='foreignkey')
    op.drop_constraint('batch_vendorcatalog_row_fk_cost', 'batch_vendorcatalog_row', type_='foreignkey')
    op.rename_table('batch_vendorcatalog_row', 'vendor_catalog_row')
    op.create_foreign_key('vendor_catalog_row_fk_batch_uuid',
                          'vendor_catalog_row', 'vendor_catalog',
                          ['batch_uuid'], ['uuid'])
    op.create_foreign_key('vendor_catalog_row_fk_product',
                          'vendor_catalog_row', 'product',
                          ['product_uuid'], ['uuid'])
    op.create_foreign_key('vendor_catalog_row_fk_cost',
                          'vendor_catalog_row', 'product_cost',
                          ['cost_uuid'], ['uuid'])



# -*- coding: utf-8 -*-
"""add dept to product batch

Revision ID: cd23a59c5d89
Revises: 14d83595afa1
Create Date: 2016-10-15 19:49:14.208311

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'cd23a59c5d89'
down_revision = u'14d83595afa1'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # "product batch rows"
    op.add_column('batch_handheld_row', sa.Column('department_name', sa.String(length=30), nullable=True))
    op.add_column('batch_handheld_row', sa.Column('department_number', sa.Integer(), nullable=True))
    op.add_column('batch_inventory_row', sa.Column('department_name', sa.String(length=30), nullable=True))
    op.add_column('batch_inventory_row', sa.Column('department_number', sa.Integer(), nullable=True))
    op.add_column('vendor_invoice_row', sa.Column('department_name', sa.String(length=30), nullable=True))
    op.add_column('vendor_invoice_row', sa.Column('department_number', sa.Integer(), nullable=True))


def downgrade():

    # "product batch rows"
    op.drop_column('vendor_invoice_row', 'department_number')
    op.drop_column('vendor_invoice_row', 'department_name')
    op.drop_column('batch_inventory_row', 'department_number')
    op.drop_column('batch_inventory_row', 'department_name')
    op.drop_column('batch_handheld_row', 'department_number')
    op.drop_column('batch_handheld_row', 'department_name')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Trainwreck model importers
"""

from __future__ import unicode_literals, absolute_import

from collections import namedtuple

from sqlalchemy import orm

from rattail import importing
from .util import ToOrFromTrainwreck


Discount = namedtuple('Discount', ['discount_type', 'description', 'amount'])


class ToTrainwreck(ToOrFromTrainwreck, importing.ToSQLAlchemy):
    """
    Base class for all Trainwreck model importers
    """
    key = 'uuid'


class TransactionImporter(ToTrainwreck):
    """
    Transaction data importer

    .. attribute:: match_on_time_field

       This time field will be used for the cache query etc.  In
       particular it controls which transactions are deemed to belong
       to a given date, which is needed when restricting the import to
       a particular date range.  Can set this to ``'upload_time'`` or
       ``'start_time'`` if necessary, depending on the nature of
       transaction data coming from the host side.
    """
    match_on_time_field = 'end_time'

    @property
    def importing_from_system(self):
        raise NotImplementedError("TODO: please define this for your subclass")

    def get_model_class(self):
        if hasattr(self, 'model_class') and self.model_class:
            return self.model_class
        trainwreck = self.config.get_trainwreck_model()
        return trainwreck.Transaction

    def cache_query(self):
        query = super(TransactionImporter, self).cache_query()
        try:
            system = self.importing_from_system
        except NotImplementedError:
            pass
        else:
            query = query.filter(self.model_class.system == system)
        time_field = getattr(self.model_class, self.match_on_time_field)
        return query.filter(time_field >= self.app.make_utc(self.start_time))\
                    .filter(time_field < self.app.make_utc(self.end_time))


class TransactionOrderMarkerImporter(ToTrainwreck):
    """
    Transaction order marker data importer

    .. attribute:: match_on_time_field

       This time field will be used for the cache query etc.  In
       particular it controls which transactions are deemed to belong
       to a given date, which is needed when restricting the import to
       a particular date range.  Can set this to ``'upload_time'`` or
       ``'start_time'`` if necessary, depending on the nature of
       transaction data coming from the host side.
    """
    match_on_time_field = 'end_time'

    @property
    def importing_from_system(self):
        raise NotImplementedError("TODO: please define this for your subclass")

    def get_model_class(self):
        if hasattr(self, 'model_class') and self.model_class:
            return self.model_class
        trainwreck = self.config.get_trainwreck_model()
        return trainwreck.TransactionOrderMarker

    @property
    def transaction_class(self):
        trainwreck = self.config.get_trainwreck_model()
        return trainwreck.Transaction

    def cache_query(self):
        query = super(TransactionOrderMarkerImporter, self).cache_query()
        trainwreck = self.config.get_trainwreck_model()

        query = query.join(trainwreck.Transaction)

        try:
            system = self.importing_from_system
        except NotImplementedError:
            pass
        else:
            query = query.filter(trainwreck.Transaction.system == system)

        time_field = getattr(trainwreck.Transaction, self.match_on_time_field)
        query = query.filter(time_field >= self.app.make_utc(self.start_time))\
                     .filter(time_field < self.app.make_utc(self.end_time))
        return query


class TransactionItemImporter(ToTrainwreck):
    """
    Transaction item data importer

    .. attribute:: match_on_time_field

       This time field will be used for the cache query etc.  In
       particular it controls which transactions are deemed to belong
       to a given date, which is needed when restricting the import to
       a particular date range.  Can set this to ``'upload_time'`` or
       ``'start_time'`` if necessary, depending on the nature of
       transaction data coming from the host side.
    """
    match_on_time_field = 'end_time'

    @property
    def supported_fields(self):
        fields = super(TransactionItemImporter, self).supported_fields
        fields = list(fields)
        fields.extend([
            'transaction_system_id',
            'discounts',
        ])
        return fields

    @property
    def importing_from_system(self):
        raise NotImplementedError("TODO: please define this for your subclass")

    def get_model_class(self):
        if hasattr(self, 'model_class') and self.model_class:
            return self.model_class
        trainwreck = self.config.get_trainwreck_model()
        return trainwreck.TransactionItem

    @property
    def transaction_class(self):
        trainwreck = self.config.get_trainwreck_model()
        return trainwreck.Transaction

    def setup(self):
        super(TransactionItemImporter, self).setup()

        if 'transaction_system_id' in self.fields:
            trainwreck = self.config.get_trainwreck_model()
            query = self.session.query(trainwreck.Transaction)
            try:
                system = self.importing_from_system
            except NotImplementedError:
                pass
            else:
                query = query.filter(trainwreck.Transaction.system == system)

            time_field = getattr(trainwreck.Transaction, self.match_on_time_field)
            query = query.filter(time_field >= self.app.make_utc(self.start_time))\
                         .filter(time_field < self.app.make_utc(self.end_time))
            self.transactions_by_system_id = self.cache_model(trainwreck.Transaction,
                                                              query=query,
                                                              key='system_id')

    def cache_query(self):
        query = super(TransactionItemImporter, self).cache_query()
        trainwreck = self.config.get_trainwreck_model()

        query = query.join(trainwreck.Transaction)

        try:
            system = self.importing_from_system
        except NotImplementedError:
            pass
        else:
            query = query.filter(trainwreck.Transaction.system == system)

        time_field = getattr(trainwreck.Transaction, self.match_on_time_field)
        query = query.filter(time_field >= self.app.make_utc(self.start_time))\
                     .filter(time_field < self.app.make_utc(self.end_time))

        if 'discounts' in self.fields:
            query = query.options(orm.joinedload(trainwreck.TransactionItem.discounts))

        return query

    def normalize_local_object(self, item):
        data = super(TransactionItemImporter, self).normalize_local_object(item)
        if not data:
            return

        if 'transaction_system_id' in self.fields:
            data['transaction_system_id'] = item.transaction.system_id

        if 'discounts' in self.fields:
            data['discounts'] = [Discount(d.discount_type, d.description, d.amount)
                                 for d in item.discounts]

        return data

    def get_transaction_by_system_id(self, system_id):
        if hasattr(self, 'transactions_by_system_id'):
            return self.transactions_by_system_id.get(system_id)

        trainwreck = self.config.get_trainwreck_model()
        query = self.session.query(trainwreck.Transaction)\
                            .filter(trainwreck.Transaction.system_id == system_id)
        try:
            system = self.importing_from_system
        except NotImplementedError:
            pass
        else:
            query = query.filter(trainwreck.Transaction.system == system)
        try:
            return query.one()
        except orm.exc.NoResultFound:
            pass

    def create_object(self, key, host_data):
        item = super(TransactionItemImporter, self).create_object(key, host_data)
        if item:

            # we may have to explicitly assign the transaction, if that uuid
            # wasn't part of our key
            # TODO: actually we do this if system_id was part of the key and do
            # not really look at uuid.  is there a better way to do or explain?
            # TODO: this also may fail outright if we get a bad system_id etc.
            if 'transaction_system_id' in self.key:
                system_id = host_data['transaction_system_id']
                txn = self.get_transaction_by_system_id(system_id)
                if not txn:
                    raise RuntimeError("transaction not found for system_id: {}".format(system_id))
                item.transaction = txn

            return item

    def update_object(self, item, host_data, local_data=None, **kwargs):
        item = super(TransactionItemImporter, self).update_object(
            item, host_data, local_data=local_data, **kwargs)
        if not item:
            return

        if 'discounts' in self.fields:
            trainwreck = self.config.get_trainwreck_model()
            if host_data['discounts']:
                if not local_data or local_data['discounts'] != host_data['discounts']:
                    # TODO: is this too heavy-handed?
                    # just rebuild the line item's "other" discount list,
                    # instead of trying to reconcile the host vs. local lists
                    if item.discounts:
                        del item.discounts[:]
                    for disc in host_data['discounts']:
                        item.discounts.append(
                            trainwreck.TransactionItemDiscount(
                                discount_type=disc.discount_type,
                                description=disc.description,
                                amount=disc.amount))
            else:
                if local_data and local_data['discounts']:
                    del item.discounts[:]
        return item


class TransactionItemDiscountImporter(ToTrainwreck):
    """
    Transaction item discount data importer

    .. attribute:: match_on_time_field

       This time field will be used for the cache query etc.  In
       particular it controls which transactions are deemed to belong
       to a given date, which is needed when restricting the import to
       a particular date range.  Can set this to ``'upload_time'`` or
       ``'start_time'`` if necessary, depending on the nature of
       transaction data coming from the host side.
    """
    match_on_time_field = 'end_time'

    @property
    def importing_from_system(self):
        raise NotImplementedError("TODO: please define this for your subclass")

    def get_model_class(self):
        if hasattr(self, 'model_class') and self.model_class:
            return self.model_class
        trainwreck = self.config.get_trainwreck_model()
        return trainwreck.TransactionItemDiscount

    def cache_query(self):
        query = super(TransactionItemDiscountImporter, self).cache_query()
        trainwreck = self.config.get_trainwreck_model()

        query = query.join(trainwreck.TransactionItem)\
                     .join(trainwreck.Transaction)

        try:
            system = self.importing_from_system
        except NotImplementedError:
            pass
        else:
            query = query.filter(trainwreck.Transaction.system == system)

        time_field = getattr(trainwreck.Transaction, self.match_on_time_field)
        query = query.filter(time_field >= self.app.make_utc(self.start_time))\
                     .filter(time_field < self.app.make_utc(self.end_time))

        return query



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Customer Order Batch Handler

Please note this is different from the Customer Order Handler.
"""

import re
import decimal
from collections import OrderedDict

import sqlalchemy as sa

from rattail.db import model
from rattail.batch import BatchHandler
from rattail.time import localtime


class CustomerOrderBatchHandler(BatchHandler):
    """
    Handler for all "customer order" batches, regardless of "mode".  The
    handler must inspect the
    :attr:`~rattail.db.model.batch.custorder.CustomerOrderBatch.mode` attribute
    of each batch it deals with, in order to determine which logic to apply.

    .. attribute:: has_custom_product_autocomplete

       If true, this flag indicates that the handler provides custom
       autocomplete logic for use when selecting a product while
       creating a new order.
    """
    batch_model_class = model.CustomerOrderBatch
    has_custom_product_autocomplete = False
    nondigits_pattern = re.compile(r'\D')

    def __init__(self, config, **kwargs):
        super(CustomerOrderBatchHandler, self).__init__(config, **kwargs)
        self.custorder_handler = self.app.get_custorder_handler()

    def init_batch(self, batch, progress=None, **kwargs):
        """
        Assign the "local" store to the batch, if applicable.
        """
        session = self.app.get_session(batch)
        batch.store = self.config.get_store(session)

    def new_order_requires_customer(self):
        """
        Returns a boolean indicating whether a *new* "customer order"
        in fact requires a proper customer account, or not.  Note that
        in all cases a new order requires a *person* to associate
        with, but technically the customer is optional, unless this
        returns true.
        """
        return self.config.getbool('rattail.custorders',
                                   'new_order_requires_customer',
                                   default=False)

    def allow_contact_info_choice(self):
        """
        Returns a boolean indicating whether the user is allowed at
        all, to choose from existing contact info options for the
        customer, vs. they just have to go with whatever the handler
        auto-provides.
        """
        return self.config.getbool('rattail.custorders',
                                   'new_orders.allow_contact_info_choice',
                                   default=False)

    def allow_contact_info_creation(self):
        """
        Returns a boolean indicating whether the user is allowed to
        enter *new* contact info for the customer.  This setting
        should only be honored if :meth:`allow_contact_info_choice()`
        also returns true.
        """
        return self.config.getbool('rattail.custorders',
                                   'new_orders.allow_contact_info_create',
                                   default=False)

    def allow_case_orders(self):
        """
        Returns a boolean indicating whether "case" orders are
        allowed.
        """
        return self.config.getbool('rattail.custorders',
                                   'allow_case_orders',
                                   default=False)

    def allow_unit_orders(self):
        """
        Returns a boolean indicating whether "unit" orders are
        allowed.
        """
        return self.config.getbool('rattail.custorders',
                                   'allow_unit_orders',
                                   default=False)

    def allow_unknown_product(self):
        """
        Returns a boolean indicating whether "unknown" products are
        allowed on new orders.
        """
        return self.config.getbool('rattail.custorders',
                                   'allow_unknown_product',
                                   default=False)

    def allow_item_discounts(self):
        """
        Returns boolean indicating whether per-item discounts are
        allowed.
        """
        return self.config.getbool('rattail.custorders',
                                   'allow_item_discounts',
                                   default=False)

    def allow_item_discounts_if_on_sale(self):
        """
        Returns boolean indicating whether per-item discounts are
        allowed when item is already on sale.
        """
        return self.config.getbool('rattail.custorders',
                                   'allow_item_discounts_if_on_sale',
                                   default=False)

    def allow_past_item_reorder(self):
        """
        Returns boolean indicating whether to expose past items for
        re-order.
        """
        return self.config.getbool('rattail.custorders',
                                   'allow_past_item_reorder',
                                   default=False)

    def product_price_may_be_questionable(self):
        """
        Returns a boolean indicating whether "any" product's price may
        be questionable.  So this isn't saying that a price *is*
        questionable but rather that it *may* be, if the user
        indicates it.  (That checkbox is only shown for the user if
        this flag is true.)
        """
        return self.config.getbool('rattail.custorders',
                                   'product_price_may_be_questionable',
                                   default=False)

    def assign_contact(self, batch, customer=None, person=None, **kwargs):
        """
        Assign the customer and/or person "contact" for the order.
        """
        clientele = self.app.get_clientele_handler()
        customer_required = self.new_order_requires_customer()

        # nb. person is always required
        if customer and not person:
            person = self.app.get_person(customer)
        if not person:
            raise ValueError("Must specify a person")

        # customer may or may not be optional
        if person and not customer:
            customer = clientele.get_customer(person)
        if customer_required and not customer:
            raise ValueError("Must specify a customer account")

        # assign contact
        batch.customer = customer
        batch.person = person

        # cache contact name
        batch.contact_name = self.get_contact_display(batch)

        # update phone/email per new contact
        batch.phone_number = None
        batch.email_address = None
        if customer_required:
            batch.phone_number = clientele.get_first_phone_number(customer)
            batch.email_address = clientele.get_first_email_address(customer)
        else:
            batch.phone_number = person.first_phone_number()
            batch.email_address = person.first_email_address()

        # always reset "add to customer" flags
        batch.clear_param('add_phone_number')
        batch.clear_param('add_email_address')

        session = self.app.get_session(batch)
        session.flush()

    def get_contact(self, batch):
        """
        Should return the contact record (i.e. Customer or Person) for
        the batch.
        """
        customer_required = self.new_order_requires_customer()

        if customer_required:
            return batch.customer
        else:
            return batch.person

    def get_contact_id(self, batch):
        """
        Should return contact ID for the batch, i.e. customer ID.
        """
        contact = self.get_contact(batch)
        if isinstance(contact, model.Customer):
            return contact.id

    def get_contact_display(self, batch):
        """
        Should return contact display text for the batch,
        i.e. customer name.
        """
        contact = self.get_contact(batch)
        if contact:
            return str(contact)

        pending = batch.pending_customer
        if pending:
            return str(pending)

    def get_contact_phones(self, batch):
        """
        Retrieve all phone records on file for the batch contact, to
        be presented as options for user to choose from when making a
        new order.
        """
        phones = []
        contact = self.get_contact(batch)
        if contact:
            phones = contact.phones

        return [self.normalize_phone(phone)
                for phone in phones]

    def normalize_phone(self, phone):
        """
        Normalize the given phone record to simple data dict, for
        passing around via JSON etc.
        """
        return {
            'uuid': phone.uuid,
            'type': phone.type,
            'number': phone.number,
            'preference': phone.preference,
            'preferred': phone.preference == 1,
        }

    def get_contact_emails(self, batch):
        """
        Retrieve all email records on file for the batch contact, to
        be presented as options for user to choose from when making a
        new order.

        Note that the default logic will exclude invalid email addresses.
        """
        emails = []
        contact = self.get_contact(batch)
        if contact:
            emails = contact.emails

        # exclude invalid
        emails = [email for email in emails
                  if not email.invalid]

        return [self.normalize_email(email)
                for email in emails]

    def normalize_email(self, email):
        """
        Normalize the given email record to simple data dict, for
        passing around via JSON etc.
        """
        return {
            'uuid': email.uuid,
            'type': email.type,
            'address': email.address,
            'invalid': email.invalid,
            'preference': email.preference,
            'preferred': email.preference == 1,
        }

    def get_contact_notes(self, batch):
        """
        Get extra "contact notes" which should be made visible to the
        user who is entering the new order.
        """
        notes = []

        invalid = False
        contact = self.get_contact(batch)
        if contact:
            invalid = [email for email in contact.emails
                       if email.invalid]
        if invalid:
            notes.append("Customer has one or more invalid email addresses on file.")

        return notes

    def unassign_contact(self, batch, **kwargs):
        """
        Unassign the customer and/or person "contact" for the order.
        """
        batch.customer = None
        batch.person = None

        # note that if batch already has a "pending" customer on file,
        # we will "restore" it as the contact info for the batch
        pending = batch.pending_customer
        if pending:
            batch.contact_name = pending.display_name
            batch.phone_number = pending.phone_number
            batch.email_address = pending.email_address
        else:
            batch.contact_name = None
            batch.phone_number = None
            batch.email_address = None

        # always reset "add to customer" flags
        batch.clear_param('add_phone_number')
        batch.clear_param('add_email_address')

        session = self.app.get_session(batch)
        session.flush()
        session.refresh(batch)

    def validate_pending_customer_data(self, batch, user, data):
        pass

    def update_pending_customer(self, batch, user, data):
        model = self.model
        people = self.app.get_people_handler()

        # first validate all data
        self.validate_pending_customer_data(batch, user, data)

        # clear out any contact it may have
        self.unassign_contact(batch)

        # create pending customer if needed
        pending = batch.pending_customer
        if not pending:
            pending = model.PendingCustomer()
            pending.user = user
            pending.status_code = self.enum.PENDING_CUSTOMER_STATUS_PENDING
            batch.pending_customer = pending

        # update pending customer info
        if 'first_name' in data:
            pending.first_name = data['first_name']
        if 'last_name' in data:
            pending.last_name = data['last_name']
        if 'display_name' in data:
            pending.display_name = data['display_name']
        else:
            pending.display_name = people.normalize_full_name(pending.first_name,
                                                              pending.last_name)
        if 'phone_number' in data:
            pending.phone_number = self.app.format_phone_number(data['phone_number'])
        if 'email_address' in data:
            pending.email_address = data['email_address']

        # also update the batch w/ contact info
        batch.contact_name = pending.display_name
        batch.phone_number = pending.phone_number
        batch.email_address = pending.email_address

    def get_case_size_for_product(self, product):
        products = self.app.get_products_handler()
        return products.get_case_size(product)

    def get_case_price_for_row(self, row):
        """
        Calculate and return the per-case price for the given row.

        NB. we do not store case price, only unit price.  maybe that
        should change some day..
        """
        if row.unit_price is not None:
            case_price = row.unit_price * (row.case_quantity or 1)
            return case_price.quantize(decimal.Decimal('0.01'))

    # TODO: this method should maybe not exist?  and caller just
    # invokes the handler directly instead?
    def customer_autocomplete(self, session, term, **kwargs):
        """
        Override the Customer autocomplete, to search by phone number
        as well as name.
        """
        autocompleter = self.app.get_autocompleter('customers.neworder')
        return autocompleter.autocomplete(session, term, **kwargs)

    # TODO: this method should maybe not exist?  and caller just
    # invokes the handler directly instead?
    def person_autocomplete(self, session, term, **kwargs):
        """
        Override the Person autocomplete, to search by phone number as
        well as name.
        """
        autocompleter = self.app.get_autocompleter('people.neworder')
        return autocompleter.autocomplete(session, term, **kwargs)

    def get_customer_info(self, batch, **kwargs):
        """
        Return a data dict containing misc. info pertaining to the
        customer/person for the order batch.
        """
        info = {
            'customer_uuid': None,
            'person_uuid': None,
            'phone_number': None,
            'email_address': None,
        }

        if batch.customer:
            info['customer_uuid'] = batch.customer.uuid
            phone = batch.customer.first_phone()
            if phone:
                info['phone_number'] = phone.number
            email = batch.customer.first_email()
            if email:
                info['email_address'] = email.address

        if batch.person:
            info['person_uuid'] = batch.person.uuid
            if not info['phone_number']:
                phone = batch.person.first_phone()
                if phone:
                    info['phone_number'] = phone.number
                email = batch.person.first_email()
                if email:
                    info['email_address'] = email.address

        return info

    def custom_product_autocomplete(self, session, term, **kwargs):
        """
        For the given term, this should return a (possibly empty) list
        of products which "match" the term.  Each element in the list
        should be a dict with "label" and "value" keys.
        """
        raise NotImplementedError("Please define the "
                                  "{}.custom_product_autocomplete() "
                                  "method.".format(__class__.__name__))

    def get_past_orders(self, batch, **kwargs):
        """
        Retrieve a list of past orders for the batch contact.
        """
        session = self.app.get_session(batch)
        model = self.model
        orders = session.query(model.CustomerOrder)

        contact = self.get_contact(batch)
        if isinstance(contact, model.Customer):
            orders = orders.filter(model.CustomerOrder.customer == contact)
        else:
            orders = orders.filter(model.CustomerOrder.person == contact)

        orders = orders.order_by(model.CustomerOrder.created.desc())
        return orders.all()

    def get_past_products(self, batch, **kwargs):
        """
        Should return a (possibly empty) list of products which have
        been ordered in the past by the customer who is associated
        with the given batch.
        """
        session = self.app.get_session(batch)
        model = self.model
        products = OrderedDict()

        # track down all order items for batch contact
        orders = self.get_past_orders(batch)
        for order in orders:
            for item in order.items:
                product = item.product
                if product:
                    # we only want the first match for each product
                    products.setdefault(product.uuid, product)

        return list(products.values())

    def get_product_info(self, batch, product, **kwargs):
        """
        Return a data dict containing misc. info pertaining to the
        given product, for the order batch.
        """
        products = self.app.get_products_handler()
        vendor = product.cost.vendor if product.cost else None
        info = {
            'uuid': product.uuid,
            'upc': str(product.upc),
            'item_id': product.item_id,
            'scancode': product.scancode,
            'upc_pretty': product.upc.pretty(),
            'brand_name': product.brand.name if product.brand else None,
            'description': product.description,
            'size': product.size,
            'full_description': product.full_description,
            'case_quantity': self.app.render_quantity(self.get_case_size_for_product(product)),
            'unit_price': float(product.regular_price.price) if product.regular_price and product.regular_price.price is not None else None,
            'unit_price_display': products.render_price(product.regular_price),
            'department_name': product.department.name if product.department else None,
            'vendor_name': vendor.name if vendor else None,
            'url': products.get_url(product),
            'image_url': products.get_image_url(product),
            'uom_choices': self.uom_choices_for_product(product),
            'default_item_discount': self.get_default_item_discount(product),
        }

        # TODO: this was somewhat copied from
        # tailbone.views.products.render_price() - should make it part
        # of the products handler instead?
        sale_price = None
        if not product.not_for_sale:
            sale_price = product.current_price
            if sale_price:
                if sale_price.price:
                    info['sale_price'] = float(sale_price.price)
                info['sale_price_display'] = products.render_price(sale_price)
                sale_ends = sale_price.ends
                if sale_ends:
                    sale_ends = localtime(self.config, sale_ends, from_utc=True).date()
                    info['sale_ends'] = str(sale_ends)
                    info['sale_ends_display'] = self.app.render_date(sale_ends)

        case_price = None
        if product.regular_price and product.regular_price is not None:
            case_size = self.get_case_size_for_product(product)
            # use sale price if there is one, else normal unit price
            unit_price = product.regular_price.price
            if sale_price:
                unit_price = sale_price.price
            case_price = (case_size or 1) * unit_price
            case_price = case_price.quantize(decimal.Decimal('0.01'))
        info['case_price'] = str(case_price) if case_price is not None else None
        info['case_price_display'] = self.app.render_currency(case_price)

        key = self.config.product_key()
        if key == 'upc':
            info['key'] = info['upc_pretty']
        else:
            info['key'] = getattr(product, key, info['upc_pretty'])

        return info

    def uom_choices_for_row(self, row):
        """
        Return a list of UOM choices for the given batch row.
        """
        if row.product:
            return self.uom_choices_for_product(row.product)

        choices = []

        # Each
        if self.allow_unit_orders():
            unit_name = self.enum.UNIT_OF_MEASURE[self.enum.UNIT_OF_MEASURE_EACH]
            choices.append({'key': self.enum.UNIT_OF_MEASURE_EACH,
                            'value': unit_name})

        # Case
        if self.allow_case_orders():
            case_text = self.enum.UNIT_OF_MEASURE[self.enum.UNIT_OF_MEASURE_CASE]
            if case_text:
                choices.append({'key': self.enum.UNIT_OF_MEASURE_CASE,
                                'value': case_text})

        return choices

    def uom_choices_for_product(self, product):
        """
        Return a list of UOM choices for the given product.
        """
        products_handler = self.app.get_products_handler()
        choices = products_handler.get_uom_choices(product)
        if not self.allow_case_orders():
            choices = [choice for choice in choices
                       if choice['key'] != self.enum.UNIT_OF_MEASURE_CASE]
        if not self.allow_unit_orders():
            choices = [choice for choice in choices
                       if choice['key'] != self.enum.UNIT_OF_MEASURE_EACH]
        return choices

    def why_not_add_product(self, product, batch):
        """
        This method can inspect the given product, and batch, to
        determine if the product may be added to the batch as a new
        row.  Useful to e.g. prevent one customer from ordering too
        many things, etc.

        :returns: If there is a reason not to add the product, should
           return that reason as a string; otherwise ``None``.
        """

    def get_default_item_discount(self, product=None, **kwargs):
        """
        Returns default item discount available.  If product is given,
        the default may be specific to its department etc.
        """
        return self.custorder_handler.get_default_item_discount(
            product=product, **kwargs)

    def add_product(self, batch, product, order_quantity, order_uom,
                    **kwargs):
        """
        Add a new row to the batch, for the given product and order
        quantity.
        """
        row = self.make_row()
        row.item_entry = product.uuid
        row.product = product
        row.order_quantity = order_quantity
        row.order_uom = order_uom

        if 'price_needs_confirmation' in kwargs:
            row.price_needs_confirmation = kwargs['price_needs_confirmation']

        if self.allow_item_discounts():
            row.discount_percent = kwargs.get('discount_percent') or 0

        self.add_row(batch, row)
        return row

    def add_pending_product(self, batch, pending_info,
                            order_quantity, order_uom,
                            **kwargs):
        """
        Add a new row to the batch, for the given "pending" product
        and order quantity.

        :param batch: Reference to the current order batch.

        :param pending_info: Dict of information about a new pending product.

        :param order_quantity: Quantity of the item to be added to the order.

        :param order_uom: Unit of measure for the order quantity.
        """
        session = self.app.get_session(batch)

        # make a new pending product
        pending_product = self.make_pending_product(**pending_info)
        session.add(pending_product)
        session.flush()

        # TODO: seems like make_pending_product() should do this?
        if pending_product.department and not pending_product.department_name:
            pending_product.department_name = pending_product.department.name

        # make a new row, w/ pending product
        row = self.make_row()
        row.pending_product = pending_product
        row.order_quantity = order_quantity
        row.order_uom = order_uom

        if self.allow_item_discounts():
            row.discount_percent = kwargs.get('discount_percent') or 0

        self.add_row(batch, row)
        return row

    def make_pending_product(self, **kwargs):
        products_handler = self.app.get_products_handler()
        if 'status_code' not in kwargs:
            kwargs['status_code'] = self.enum.PENDING_PRODUCT_STATUS_PENDING
        return products_handler.make_pending_product(**kwargs)

    def update_pending_product(self, row, data):
        """
        Update the pending product data for the given batch row.
        """
        # create pending product if needed
        pending = row.pending_product
        if not pending:
            pending = self.make_pending_product(**data)

        simple_fields = [
            'scancode',
            'item_id',
            'item_type',
            'department_name',
            'department_uuid',
            'brand_name',
            'brand_uuid',
            'description',
            'size',
            'vendor_name',
            'vendor_uuid',
            'vendor_item_code',
            'special_order',
            'notes',
        ]

        decimal_fields = [
            'unit_cost',
            'case_size',
            'case_cost',
            'regular_price_amount',
        ]

        # update pending product info
        if 'upc' in data:
            pending.upc = self.app.make_gpc(data['upc']) if data['upc'] else None
        for field in simple_fields:
            if field in data:
                setattr(pending, field, data[field])
        for field in decimal_fields:
            if field in data:
                value = data[field]
                value = decimal.Decimal(value) if value is not None else None
                setattr(pending, field, value)

        # refresh the row per new pending product info
        self.refresh_row(row)

    def refresh_row(self, row):
        if not row.product:
            if row.item_entry:
                session = self.app.get_session(row)
                product = self.locate_product_for_entry(session, row.item_entry)
                if product:
                    row.product = product
            if not row.product and not row.pending_product:
                row.status_code = row.STATUS_PRODUCT_NOT_FOUND
                return

        product = row.product
        if product:

            row.product_upc = product.upc
            row.product_item_id = product.item_id
            row.product_scancode = product.scancode
            row.product_brand = str(product.brand or "")
            row.product_description = product.description
            row.product_size = product.size
            row.product_weighed = product.weighed
            row.case_quantity = self.get_case_size_for_product(product)

            department = product.department
            row.department_number = department.number if department else None
            row.department_name = department.name if department else None

            cost = product.cost
            row.product_unit_cost = cost.unit_cost if cost else None

            regprice = product.regular_price
            row.unit_regular_price = regprice.price if regprice else None

            curprice = product.current_price
            row.unit_sale_price = curprice.price if curprice else None
            row.sale_ends = curprice.ends if curprice else None

            row.unit_price = row.unit_sale_price or row.unit_regular_price

        else:
            self.refresh_row_from_pending_product(row)

        # we need to know if total price is updated
        old_total = row.total_price

        # maybe update total price
        if row.unit_price is None:
            row.total_price = None
        elif not row.unit_price:
            row.total_price = 0
        else:
            row.total_price = row.unit_price * row.order_quantity
            if row.order_uom == self.enum.UNIT_OF_MEASURE_CASE:
                row.total_price *= (row.case_quantity or 1)
            if row.discount_percent:
                row.total_price = (float(row.total_price)
                                   * (100 - float(row.discount_percent))
                                   / 100.0)
            row.total_price = decimal.Decimal('{:0.2f}'.format(row.total_price))

        # update total price for batch too, if it changed
        if row.total_price != old_total:
            batch = row.batch
            batch.total_price = ((batch.total_price or 0)
                                 + (row.total_price or 0)
                                 - (old_total or 0))

        if not row.product and row.pending_product:
            row.status_code = row.STATUS_PENDING_PRODUCT
        else:
            row.status_code = row.STATUS_OK

    def refresh_row_from_pending_product(self, row):
        """
        Refresh basic row attributes from its pending product.
        """
        pending = row.pending_product
        if not pending:
            return

        row.product_upc = pending.upc
        row.product_item_id = pending.item_id
        row.product_scancode = pending.scancode
        row.product_brand = str(pending.brand or pending.brand_name or '')
        row.product_description = pending.description
        row.product_size = pending.size
        # TODO: is this even important?  pending does not have it
        # row.product_weighed = pending.weighed
        row.case_quantity = pending.case_size

        if pending.department:
            row.department = pending.department
            row.department_number = pending.department.number
            row.department_name = pending.department.name
        elif pending.department_name:
            row.department = None
            row.department_number = None
            row.department_name = pending.department_name
        else:
            row.department = None
            row.department_number = None
            row.department_name = None

        row.product_unit_cost = pending.unit_cost
        row.unit_regular_price = pending.regular_price_amount
        row.unit_price = row.unit_regular_price
        row.unit_sale_price = None
        row.sale_ends = None

    def remove_row(self, row):
        batch = row.batch

        if not row.removed:
            row.removed = True

            if row.total_price:
                batch.total_price = (batch.total_price or 0) - row.total_price

        self.refresh_batch_status(batch)

    def delete_extra_data(self, batch, progress=None, **kwargs):

        # do the normal stuff (e.g. delete input files)
        super(CustomerOrderBatchHandler, self).delete_extra_data(
            batch, progress=progress, **kwargs)

        session = self.app.get_session(batch)

        # delete pending customer if present
        pending = batch.pending_customer
        if pending:
            batch.pending_customer = None
            session.delete(pending)

        # delete any pending products if present
        for row in batch.data_rows:
            pending = row.pending_product
            if pending:
                row.pending_product = None
                session.delete(pending)

    def execute(self, batch, user=None, progress=None, **kwargs):
        """
        Default behavior here will create and return a new rattail
        Customer Order.  It also may "add contact info" e.g. to the
        customer record.  Override as needed.
        """
        rows = batch.active_rows()
        order = self.make_new_order(batch, user=user, progress=progress, **kwargs)
        self.update_contact_info(batch, user)
        self.mark_pending_things_ready(batch, rows)
        return order

    def mark_pending_things_ready(self, batch, rows):
        if batch.pending_customer:
            batch.pending_customer.status = self.enum.PENDING_CUSTOMER_STATUS_READY
        for row in rows:
            if row.pending_product:
                row.pending_product.status = self.enum.PENDING_PRODUCT_STATUS_READY

    def update_contact_info(self, batch, user, **kwargs):
        """
        Update contact info from the batch, onto the customer record.
        """
        if batch.get_param('add_phone_number'):
            self.add_phone_number(batch, user)
        if batch.get_param('add_email_address'):
            self.add_email_address(batch, user)

    def add_phone_number(self, batch, user, **kwargs):
        """
        Add phone number from the batch to the customer record.

        Note that the default behavior does *not* do that, but instead
        will send an email alert to configured recipient(s) with the
        update request.
        """
        self.app.send_email('new_phone_requested', {
            'user': user,
            'user_display': user.display_name if user else "(unknown user)",
            'contact': self.get_contact(batch),
            'contact_id': self.get_contact_id(batch),
            'phone_number': batch.phone_number,
        })

    def add_email_address(self, batch, user, **kwargs):
        """
        Add email address from the batch to the customer record.

        Note that the default behavior does *not* do that, but instead
        will send an email alert to configured recipient(s) with the
        update request.
        """
        self.app.send_email('new_email_requested', {
            'user': user,
            'user_display': user.display_name if user else "(unknown user)",
            'contact': self.get_contact(batch),
            'contact_id': self.get_contact_id(batch),
            'email_address': batch.email_address,
        })

    def make_new_order(self, batch, user=None, progress=None, **kwargs):
        """
        Create and return a new rattail Customer Order based on the
        batch contents.
        """
        batch_fields = [
            'store',
            'id',
            'customer',
            'person',
            'pending_customer',
            'phone_number',
            'email_address',
            'total_price',
        ]

        order = model.CustomerOrder()
        order.created_by = user
        order.status_code = self.enum.CUSTORDER_STATUS_ORDERED
        for field in batch_fields:
            setattr(order, field, getattr(batch, field))

        session = self.app.get_session(batch)
        session.add(order)
        session.flush()

        row_fields = [
            'product',
            'pending_product_uuid',
            'product_upc',
            'product_brand',
            'product_description',
            'product_size',
            'product_weighed',
            'department_number',
            'department_name',
            'case_quantity',
            'order_quantity',
            'order_uom',
            'product_unit_cost',
            'unit_price',
            'discount_percent',
            'total_price',
            'price_needs_confirmation',
            'paid_amount',
            'payment_transaction_number',
        ]

        def convert(row, i):

            # add new order item
            item = model.CustomerOrderItem()
            item.sequence = i
            for field in row_fields:
                setattr(item, field, getattr(row, field))
            order.items.append(item)

            # set initial status and attach events
            self.set_initial_item_status(item, user)

        self.progress_loop(convert, batch.active_rows(), progress,
                           message="Converting batch rows to order items")

        session.flush()
        return order

    def set_initial_item_status(self, item, user, **kwargs):
        """
        Set the initial status for the given order item, and attach
        any events.

        The first logical status is ``CUSTORDER_ITEM_STATUS_INITIATED``
        and an item may stay there if there is some other step(s)
        which must occur before the item is ready to proceed.  For
        instance the default logic will leave it there if the price
        needs to be confirmed, but you can override as needed, for
        instance if you require payment up-front.

        The second status is ``CUSTORDER_ITEM_STATUS_READY`` which
        indicates the item is ready to proceed.  The default logic
        will auto-advance the item to this status if the price does
        *not* need to be confirmed.  Again you may need to override
        e.g. to prevent this until up-front payment is received.
        """
        # set "initiated" status
        item.status_code = self.enum.CUSTORDER_ITEM_STATUS_INITIATED
        item.add_event(self.enum.CUSTORDER_ITEM_EVENT_INITIATED, user)

        # but if the price is good...
        if not item.price_needs_confirmation:

            # then we set "ready" status
            item.status_code = self.enum.CUSTORDER_ITEM_STATUS_READY
            item.status_text = "everything looks normal"

            item.add_event(self.enum.CUSTORDER_ITEM_EVENT_READY, user,
                           note=item.status_text)



# -*- coding: utf-8 -*-
"""add versioning

Revision ID: 47d30c955111
Revises: 33b76c3cb892
Create Date: 2015-02-07 12:55:40.293606

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '47d30c955111'
down_revision = u'33b76c3cb892'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types


def upgrade():

    connection = op.get_bind()

    # transaction
    op.create_table('transaction',
                    sa.Column('issued_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.BigInteger(), nullable=False),
                    sa.Column('remote_addr', sa.String(length=50), nullable=True),
                    sa.Column('user_uuid', sa.String(length=32), nullable=True),
                    sa.ForeignKeyConstraint(['user_uuid'], ['user.uuid'], ),
                    sa.PrimaryKeyConstraint('id')
                    )
    op.create_index(op.f('ix_transaction_user_uuid'), 'transaction', ['user_uuid'], unique=False)

    # email
    op.create_table('email_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('parent_type', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('parent_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('preference', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('type', sa.String(length=15), autoincrement=False, nullable=True),
                    sa.Column('address', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_email_version_end_transaction_id'), 'email_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_email_version_operation_type'), 'email_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_email_version_transaction_id'), 'email_version', ['transaction_id'], unique=False)

    # phone
    op.create_table('phone_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('parent_type', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('parent_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('preference', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('type', sa.String(length=15), autoincrement=False, nullable=True),
                    sa.Column('number', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_phone_version_end_transaction_id'), 'phone_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_phone_version_operation_type'), 'phone_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_phone_version_transaction_id'), 'phone_version', ['transaction_id'], unique=False)

    # person
    op.create_table('person_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('first_name', sa.String(length=50), autoincrement=False, nullable=True),
                    sa.Column('last_name', sa.String(length=50), autoincrement=False, nullable=True),
                    sa.Column('display_name', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_person_version_end_transaction_id'), 'person_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_person_version_operation_type'), 'person_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_person_version_transaction_id'), 'person_version', ['transaction_id'], unique=False)

    # role
    op.create_table('role_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('name', sa.String(length=25), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_role_version_end_transaction_id'), 'role_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_role_version_operation_type'), 'role_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_role_version_transaction_id'), 'role_version', ['transaction_id'], unique=False)

    # user
    op.create_table('user_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('username', sa.String(length=25), autoincrement=False, nullable=True),
                    sa.Column('person_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('active', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_user_version_end_transaction_id'), 'user_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_user_version_operation_type'), 'user_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_user_version_transaction_id'), 'user_version', ['transaction_id'], unique=False)

    # store
    op.create_table('store_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.String(length=10), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('database_key', sa.String(length=30), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_store_version_end_transaction_id'), 'store_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_store_version_operation_type'), 'store_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_store_version_transaction_id'), 'store_version', ['transaction_id'], unique=False)

    # employee
    op.create_table('employee_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('person_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('status', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('display_name', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_employee_version_end_transaction_id'), 'employee_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_employee_version_operation_type'), 'employee_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_employee_version_transaction_id'), 'employee_version', ['transaction_id'], unique=False)

    # customer_group
    op.create_table('customer_group_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_customer_group_version_end_transaction_id'), 'customer_group_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_customer_group_version_operation_type'), 'customer_group_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_customer_group_version_transaction_id'), 'customer_group_version', ['transaction_id'], unique=False)

    # customer
    op.create_table('customer_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('email_preference', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_customer_version_end_transaction_id'), 'customer_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_customer_version_operation_type'), 'customer_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_customer_version_transaction_id'), 'customer_version', ['transaction_id'], unique=False)

    # vendor
    op.create_table('vendor_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.String(length=15), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=50), autoincrement=False, nullable=True),
                    sa.Column('special_discount', sa.Numeric(precision=5, scale=3), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_vendor_version_end_transaction_id'), 'vendor_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_vendor_version_operation_type'), 'vendor_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_vendor_version_transaction_id'), 'vendor_version', ['transaction_id'], unique=False)

    # department
    op.create_table('department_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('number', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=30), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_department_version_end_transaction_id'), 'department_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_department_version_operation_type'), 'department_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_department_version_transaction_id'), 'department_version', ['transaction_id'], unique=False)

    # subdepartment
    op.create_table('subdepartment_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('number', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=30), autoincrement=False, nullable=True),
                    sa.Column('department_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_subdepartment_version_end_transaction_id'), 'subdepartment_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_subdepartment_version_operation_type'), 'subdepartment_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_subdepartment_version_transaction_id'), 'subdepartment_version', ['transaction_id'], unique=False)

    # category
    op.create_table('category_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('number', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=50), autoincrement=False, nullable=True),
                    sa.Column('department_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_category_version_end_transaction_id'), 'category_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_category_version_operation_type'), 'category_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_category_version_transaction_id'), 'category_version', ['transaction_id'], unique=False)

    # family
    op.create_table('family_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('code', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=50), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_family_version_end_transaction_id'), 'family_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_family_version_operation_type'), 'family_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_family_version_transaction_id'), 'family_version', ['transaction_id'], unique=False)

    # report_code
    op.create_table('report_code_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('code', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=50), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_report_code_version_end_transaction_id'), 'report_code_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_report_code_version_operation_type'), 'report_code_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_report_code_version_transaction_id'), 'report_code_version', ['transaction_id'], unique=False)

    # brand
    op.create_table('brand_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('name', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_brand_version_end_transaction_id'), 'brand_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_brand_version_operation_type'), 'brand_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_brand_version_transaction_id'), 'brand_version', ['transaction_id'], unique=False)

    # product
    op.create_table('product_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('upc', rattail.db.types.GPCType(), autoincrement=False, nullable=True),
                    sa.Column('department_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('subdepartment_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('category_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('family_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('report_code_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('brand_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('description', sa.String(length=60), autoincrement=False, nullable=True),
                    sa.Column('description2', sa.String(length=60), autoincrement=False, nullable=True),
                    sa.Column('size', sa.String(length=30), autoincrement=False, nullable=True),
                    sa.Column('unit_of_measure', sa.String(length=4), autoincrement=False, nullable=True),
                    sa.Column('not_for_sale', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('regular_price_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('current_price_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_product_version_end_transaction_id'), 'product_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_product_version_operation_type'), 'product_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_product_version_transaction_id'), 'product_version', ['transaction_id'], unique=False)

    # product_code
    op.create_table('product_code_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('product_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('ordinal', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('code', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_product_code_version_end_transaction_id'), 'product_code_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_product_code_version_operation_type'), 'product_code_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_product_code_version_transaction_id'), 'product_code_version', ['transaction_id'], unique=False)

    # product_cost
    op.create_table('product_cost_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('product_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('vendor_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('preference', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('code', sa.String(length=15), autoincrement=False, nullable=True),
                    sa.Column('case_size', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('case_cost', sa.Numeric(precision=9, scale=5), autoincrement=False, nullable=True),
                    sa.Column('pack_size', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('pack_cost', sa.Numeric(precision=9, scale=5), autoincrement=False, nullable=True),
                    sa.Column('unit_cost', sa.Numeric(precision=9, scale=5), autoincrement=False, nullable=True),
                    sa.Column('effective', sa.DateTime(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_product_cost_version_end_transaction_id'), 'product_cost_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_product_cost_version_operation_type'), 'product_cost_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_product_cost_version_transaction_id'), 'product_cost_version', ['transaction_id'], unique=False)

    # product_price
    op.create_table('product_price_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('product_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('type', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('level', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('starts', sa.DateTime(), autoincrement=False, nullable=True),
                    sa.Column('ends', sa.DateTime(), autoincrement=False, nullable=True),
                    sa.Column('price', sa.Numeric(precision=8, scale=3), autoincrement=False, nullable=True),
                    sa.Column('multiple', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('pack_price', sa.Numeric(precision=8, scale=3), autoincrement=False, nullable=True),
                    sa.Column('pack_multiple', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_product_price_version_end_transaction_id'), 'product_price_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_product_price_version_operation_type'), 'product_price_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_product_price_version_transaction_id'), 'product_price_version', ['transaction_id'], unique=False)

    # label_profile
    op.create_table('label_profile_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('ordinal', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('code', sa.String(length=3), autoincrement=False, nullable=True),
                    sa.Column('description', sa.String(length=50), autoincrement=False, nullable=True),
                    sa.Column('printer_spec', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('formatter_spec', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('format', sa.Text(), autoincrement=False, nullable=True),
                    sa.Column('visible', sa.Boolean(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_label_profile_version_end_transaction_id'), 'label_profile_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_label_profile_version_operation_type'), 'label_profile_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_label_profile_version_transaction_id'), 'label_profile_version', ['transaction_id'], unique=False)


def downgrade():

    # label_profile
    op.drop_index(op.f('ix_label_profile_version_transaction_id'), table_name='label_profile_version')
    op.drop_index(op.f('ix_label_profile_version_operation_type'), table_name='label_profile_version')
    op.drop_index(op.f('ix_label_profile_version_end_transaction_id'), table_name='label_profile_version')
    op.drop_table('label_profile_version')

    # product_price
    op.drop_index(op.f('ix_product_price_version_transaction_id'), table_name='product_price_version')
    op.drop_index(op.f('ix_product_price_version_operation_type'), table_name='product_price_version')
    op.drop_index(op.f('ix_product_price_version_end_transaction_id'), table_name='product_price_version')
    op.drop_table('product_price_version')

    # product_cost
    op.drop_index(op.f('ix_product_cost_version_transaction_id'), table_name='product_cost_version')
    op.drop_index(op.f('ix_product_cost_version_operation_type'), table_name='product_cost_version')
    op.drop_index(op.f('ix_product_cost_version_end_transaction_id'), table_name='product_cost_version')
    op.drop_table('product_cost_version')

    # product_code
    op.drop_index(op.f('ix_product_code_version_transaction_id'), table_name='product_code_version')
    op.drop_index(op.f('ix_product_code_version_operation_type'), table_name='product_code_version')
    op.drop_index(op.f('ix_product_code_version_end_transaction_id'), table_name='product_code_version')
    op.drop_table('product_code_version')

    # product
    op.drop_index(op.f('ix_product_version_transaction_id'), table_name='product_version')
    op.drop_index(op.f('ix_product_version_operation_type'), table_name='product_version')
    op.drop_index(op.f('ix_product_version_end_transaction_id'), table_name='product_version')
    op.drop_table('product_version')

    # brand
    op.drop_index(op.f('ix_brand_version_transaction_id'), table_name='brand_version')
    op.drop_index(op.f('ix_brand_version_operation_type'), table_name='brand_version')
    op.drop_index(op.f('ix_brand_version_end_transaction_id'), table_name='brand_version')
    op.drop_table('brand_version')

    # report_code
    op.drop_index(op.f('ix_report_code_version_transaction_id'), table_name='report_code_version')
    op.drop_index(op.f('ix_report_code_version_operation_type'), table_name='report_code_version')
    op.drop_index(op.f('ix_report_code_version_end_transaction_id'), table_name='report_code_version')
    op.drop_table('report_code_version')

    # family
    op.drop_index(op.f('ix_family_version_transaction_id'), table_name='family_version')
    op.drop_index(op.f('ix_family_version_operation_type'), table_name='family_version')
    op.drop_index(op.f('ix_family_version_end_transaction_id'), table_name='family_version')
    op.drop_table('family_version')

    # category
    op.drop_index(op.f('ix_category_version_transaction_id'), table_name='category_version')
    op.drop_index(op.f('ix_category_version_operation_type'), table_name='category_version')
    op.drop_index(op.f('ix_category_version_end_transaction_id'), table_name='category_version')
    op.drop_table('category_version')

    # subdepartment
    op.drop_index(op.f('ix_subdepartment_version_transaction_id'), table_name='subdepartment_version')
    op.drop_index(op.f('ix_subdepartment_version_operation_type'), table_name='subdepartment_version')
    op.drop_index(op.f('ix_subdepartment_version_end_transaction_id'), table_name='subdepartment_version')
    op.drop_table('subdepartment_version')

    # department
    op.drop_index(op.f('ix_department_version_transaction_id'), table_name='department_version')
    op.drop_index(op.f('ix_department_version_operation_type'), table_name='department_version')
    op.drop_index(op.f('ix_department_version_end_transaction_id'), table_name='department_version')
    op.drop_table('department_version')

    # vendor
    op.drop_index(op.f('ix_vendor_version_transaction_id'), table_name='vendor_version')
    op.drop_index(op.f('ix_vendor_version_operation_type'), table_name='vendor_version')
    op.drop_index(op.f('ix_vendor_version_end_transaction_id'), table_name='vendor_version')
    op.drop_table('vendor_version')

    # customer
    op.drop_index(op.f('ix_customer_version_transaction_id'), table_name='customer_version')
    op.drop_index(op.f('ix_customer_version_operation_type'), table_name='customer_version')
    op.drop_index(op.f('ix_customer_version_end_transaction_id'), table_name='customer_version')
    op.drop_table('customer_version')

    # customer_group
    op.drop_index(op.f('ix_customer_group_version_transaction_id'), table_name='customer_group_version')
    op.drop_index(op.f('ix_customer_group_version_operation_type'), table_name='customer_group_version')
    op.drop_index(op.f('ix_customer_group_version_end_transaction_id'), table_name='customer_group_version')
    op.drop_table('customer_group_version')

    # employee
    op.drop_index(op.f('ix_employee_version_transaction_id'), table_name='employee_version')
    op.drop_index(op.f('ix_employee_version_operation_type'), table_name='employee_version')
    op.drop_index(op.f('ix_employee_version_end_transaction_id'), table_name='employee_version')
    op.drop_table('employee_version')

    # store
    op.drop_index(op.f('ix_store_version_transaction_id'), table_name='store_version')
    op.drop_index(op.f('ix_store_version_operation_type'), table_name='store_version')
    op.drop_index(op.f('ix_store_version_end_transaction_id'), table_name='store_version')
    op.drop_table('store_version')

    # user
    op.drop_index(op.f('ix_user_version_transaction_id'), table_name='user_version')
    op.drop_index(op.f('ix_user_version_operation_type'), table_name='user_version')
    op.drop_index(op.f('ix_user_version_end_transaction_id'), table_name='user_version')
    op.drop_table('user_version')

    # role
    op.drop_index(op.f('ix_role_version_transaction_id'), table_name='role_version')
    op.drop_index(op.f('ix_role_version_operation_type'), table_name='role_version')
    op.drop_index(op.f('ix_role_version_end_transaction_id'), table_name='role_version')
    op.drop_table('role_version')

    # person
    op.drop_index(op.f('ix_person_version_transaction_id'), table_name='person_version')
    op.drop_index(op.f('ix_person_version_operation_type'), table_name='person_version')
    op.drop_index(op.f('ix_person_version_end_transaction_id'), table_name='person_version')
    op.drop_table('person_version')

    # phone
    op.drop_index(op.f('ix_phone_version_transaction_id'), table_name='phone_version')
    op.drop_index(op.f('ix_phone_version_operation_type'), table_name='phone_version')
    op.drop_index(op.f('ix_phone_version_end_transaction_id'), table_name='phone_version')
    op.drop_table('phone_version')

    # email
    op.drop_index(op.f('ix_email_version_transaction_id'), table_name='email_version')
    op.drop_index(op.f('ix_email_version_operation_type'), table_name='email_version')
    op.drop_index(op.f('ix_email_version_end_transaction_id'), table_name='email_version')
    op.drop_table('email_version')

    # transaction
    op.drop_index(op.f('ix_transaction_user_uuid'), table_name='transaction')
    op.drop_table('transaction')



# -*- coding: utf-8; -*-
"""add vendor_catalog.future

Revision ID: 0174642e7b60
Revises: d3b6e3c971b4
Create Date: 2018-04-09 22:36:01.163993

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '0174642e7b60'
down_revision = u'd3b6e3c971b4'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # vendor_catalog
    op.add_column('vendor_catalog', sa.Column('future', sa.Boolean(), nullable=True))
    op.add_column('vendor_catalog_row', sa.Column('ends', sa.DateTime(), nullable=True))
    op.add_column('vendor_catalog_row', sa.Column('starts', sa.DateTime(), nullable=True))


def downgrade():

    # vendor_catalog
    op.drop_column('vendor_catalog_row', 'starts')
    op.drop_column('vendor_catalog_row', 'ends')
    op.drop_column('vendor_catalog', 'future')



# -*- coding: utf-8; -*-
"""make case_size decimal

Revision ID: 4594843ad852
Revises: 90e3b9ea5356
Create Date: 2022-12-22 16:14:36.139342

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '4594843ad852'
down_revision = '90e3b9ea5356'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_vendorcatalog_row
    op.alter_column('batch_vendorcatalog_row', 'case_size', type_=sa.Numeric(precision=9, scale=4))
    op.alter_column('batch_vendorcatalog_row', 'old_case_size', type_=sa.Numeric(precision=9, scale=4))

    # batch_vendorinvoice_row
    op.alter_column('batch_vendorinvoice_row', 'case_quantity', type_=sa.Numeric(precision=9, scale=4))


def downgrade():

    # batch_vendorinvoice_row
    op.alter_column('batch_vendorinvoice_row', 'case_quantity', type_=sa.Integer())

    # batch_vendorcatalog_row
    op.alter_column('batch_vendorcatalog_row', 'case_size', type_=sa.Integer())
    op.alter_column('batch_vendorcatalog_row', 'old_case_size', type_=sa.Integer())



# -*- coding: utf-8; -*-
"""add vendor.abbreviation

Revision ID: 05d1ed3a4a28
Revises: 765ce0ae5bbd
Create Date: 2018-06-01 14:28:56.127658

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '05d1ed3a4a28'
down_revision = '765ce0ae5bbd'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # vendor
    op.add_column('vendor', sa.Column('abbreviation', sa.String(length=20), nullable=True))
    op.add_column('vendor_version', sa.Column('abbreviation', sa.String(length=20), autoincrement=False, nullable=True))


def downgrade():

    # vendor
    op.drop_column('vendor_version', 'abbreviation')
    op.drop_column('vendor', 'abbreviation')



# -*- coding: utf-8 -*-
"""make message recips unique

Revision ID: 33c3b9acc341
Revises: 6e6d6e2ef0e
Create Date: 2016-02-11 17:42:19.561486

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '33c3b9acc341'
down_revision = u'6e6d6e2ef0e'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # message_recip
    op.create_unique_constraint(u'message_recip_uq_message_recipient', 'message_recip', ['message_uuid', 'recipient_uuid'])


def downgrade():

    # message_recip
    op.drop_constraint(u'message_recip_uq_message_recipient', 'message_recip', type_='unique')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendor Invoices
"""

from __future__ import unicode_literals, absolute_import

import os
from decimal import Decimal

import six

from rattail.exceptions import RattailError
from rattail.util import load_entry_points
from rattail.files import resource_path
from rattail.excel import ExcelReaderXLSX


class InvoiceParser(object):
    """
    Base class for all vendor invoice parsers.
    """

    def __init__(self, config):
        self.config = config
        self.app = config.get_app()
        self.enum = config.get_enum()
        self.model = config.get_model()

    @property
    def key(self):
        """
        Key for the parser.  Must be unique among all invoice parsers.
        """
        raise NotImplementedError("Invoice parser has no key: {0}".format(repr(self)))

    @property
    def vendor_key(self):
        """
        Key for the vendor.  This key will be used to locate an entry in the
        settings table, e.g. ``'rattail.vendor.unfi'`` for a key of ``'unfi'``.
        The value of this setting must be an exact match to either a
        :attr:`rattail.db.model.Vendor.uuid` or
        :attr:`rattail.db.model.Vendor.id` within the system.
        """
        raise NotImplementedError

    def get_vendor(self, session):
        """
        Fetch the :class:~rattail.db.model.vendors.Vendor` record
        which is associated with the current parser, if any.
        """
        if self.vendor_key:
            vendor_handler = self.app.get_vendor_handler()
            return vendor_handler.get_vendor(session, self.vendor_key)

    def parse_invoice_date(self, path):
        """
        Parse the invoice date from the invoice file.
        """

    def parse_invoice_number(self, path):
        """
        Parse the invoice number from the invoice file.
        """

    def parse_invoice_total(self, path):
        """
        Parse the invoice total from the invoice file.
        """

    def parse_rows(self, data_path):
        """
        Parse the given data file, returning all rows found within it.
        """
        raise NotImplementedError("Invoice parser has no `parse_rows()` method: {0}".format(repr(self.key)))

    def make_row(self):
        """
        Make a new, empty row, to be populated from a parsed invoice
        line item.
        """
        from rattail.db import model

        # TODO: should use some other class that does not involve DB
        row = model.VendorInvoiceBatchRow()
        return row

    def decimal(self, value, scale=4):
        """
        Convert a value to a decimal.
        """
        if value is None:
            return
        # No reason to convert integers, really.
        if isinstance(value, (Decimal, int)):
            return value
        if isinstance(value, float):
            return Decimal("{{:0.{}f}}".format(scale).format(value))
        value = value.strip()
        if value:
            return Decimal(value)

    def int_(self, value):
        """
        Convert a value to an integer.
        """
        value = value.strip() or 0
        return int(value)


class ExcelInvoiceParser(InvoiceParser):
    """
    Base class for Excel vendor invoice parsers.
    """

    def get_excel_reader(self, path):
        """
        Return a :class:`~rattail.excel:ExcelReaderXLXS` instance for
        the given path.
        """
        if not hasattr(self, 'excel_reader'):
            kwargs = self.get_excel_reader_kwargs()
            self.excel_reader = ExcelReaderXLSX(path, **kwargs)
        return self.excel_reader

    def get_excel_reader_kwargs(self, **kwargs):
        return kwargs


class PDFInvoiceParser(InvoiceParser):
    """
    Base class for PDF vendor invoice parsers.
    """

    def parse_invoice_date(self, path):
        data = self.extract_pdf_data(path)
        return data['date']

    def parse_invoice_number(self, path):
        data = self.extract_pdf_data(path)
        return data['invoice_number']

    def parse_invoice_total(self, path):
        data = self.extract_pdf_data(path)
        return data['amount']

    def extract_pdf_data(self, path, template_dir=None):
        """
        Invoke the ``invoice2data`` library to automatically extract
        data from the given PDF file.
        """
        if not hasattr(self, 'extracted_pdf_data'):
            from invoice2data import extract_data
            from invoice2data.extract.loader import read_templates

            if not template_dir:
                template_dir = self.config.require('rattail.invoice2data',
                                                   'templates')
            template_dir = resource_path(template_dir)
            templates = read_templates(template_dir)
            result = extract_data(path, templates=templates)
            if not result:
                raise RuntimeError("could not find a matching template for the invoice!")
            self.extracted_pdf_data = result

        return self.extracted_pdf_data


@six.python_2_unicode_compatible
class InvoiceParserNotFound(RattailError):
    """
    Exception raised when a vendor invoice parser is required, but cannot be
    located.
    """

    def __init__(self, key):
        self.key = key

    def __str__(self):
        return "Vendor invoice parser with key {} cannot be located.".format(self.key)


def get_invoice_parsers():
    """
    Returns a dictionary of installed vendor invoice parser classes.
    """
    return load_entry_points('rattail.vendors.invoices.parsers')


def get_invoice_parser(config, key):
    """
    Fetch a vendor invoice parser by key.  If the parser class can be located,
    this will return an instance thereof; otherwise returns ``None``.
    """
    parser = get_invoice_parsers().get(key)
    if parser:
        return parser(config)
    return None


def require_invoice_parser(config, key):
    """
    Fetch a vendor invoice parser by key.  If the parser class can be located,
    this will return an instance thereof; otherwise raises an exception.
    """
    parser = get_invoice_parser(config, key)
    if not parser:
        raise InvoiceParserNotFound(key)
    return parser


def iter_invoice_parsers():
    """
    Returns an iterator over the installed vendor invoice parsers.
    """
    parsers = get_invoice_parsers()
    return parsers.values()



# -*- coding: utf-8 -*-
"""add pricing batch

Revision ID: 3f266c89a3d4
Revises: 20ec527fad0a
Create Date: 2016-11-21 18:01:48.938952

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '3f266c89a3d4'
down_revision = u'20ec527fad0a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # batch_pricing
    op.create_table('batch_pricing',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.Column('cognized', sa.DateTime(), nullable=True),
                    sa.Column('cognized_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('rowcount', sa.Integer(), nullable=True),
                    sa.Column('executed', sa.DateTime(), nullable=True),
                    sa.Column('executed_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('purge', sa.Date(), nullable=True),
                    sa.ForeignKeyConstraint(['cognized_by_uuid'], [u'user.uuid'], name=u'batch_pricing_fk_cognized_by'),
                    sa.ForeignKeyConstraint(['created_by_uuid'], [u'user.uuid'], name=u'batch_pricing_fk_created_by'),
                    sa.ForeignKeyConstraint(['executed_by_uuid'], [u'user.uuid'], name=u'batch_pricing_fk_executed_by'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # batch_pricing_row
    op.create_table('batch_pricing_row',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('batch_uuid', sa.String(length=32), nullable=False),
                    sa.Column('sequence', sa.Integer(), nullable=False),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.Column('status_text', sa.String(length=255), nullable=True),
                    sa.Column('removed', sa.Boolean(), nullable=False),
                    sa.Column('upc', rattail.db.types.GPCType(), nullable=True),
                    sa.Column('product_uuid', sa.String(length=32), nullable=True),
                    sa.Column('brand_name', sa.String(length=100), nullable=True),
                    sa.Column('description', sa.String(length=255), nullable=True),
                    sa.Column('size', sa.String(length=255), nullable=True),
                    sa.Column('department_number', sa.Integer(), nullable=True),
                    sa.Column('department_name', sa.String(length=30), nullable=True),
                    sa.Column('regular_unit_cost', sa.Numeric(precision=9, scale=5), nullable=True),
                    sa.Column('discounted_unit_cost', sa.Numeric(precision=9, scale=5), nullable=True),
                    sa.Column('old_price', sa.Numeric(precision=8, scale=3), nullable=True),
                    sa.Column('price_margin', sa.Numeric(precision=5, scale=3), nullable=True),
                    sa.Column('price_markup', sa.Numeric(precision=4, scale=2), nullable=True),
                    sa.Column('new_price', sa.Numeric(precision=8, scale=3), nullable=True),
                    sa.Column('price_diff', sa.Numeric(precision=8, scale=3), nullable=True),
                    sa.ForeignKeyConstraint(['batch_uuid'], [u'batch_pricing.uuid'], name=u'batch_pricing_row_fk_batch'),
                    sa.ForeignKeyConstraint(['product_uuid'], [u'product.uuid'], name=u'batch_pricing_row_fk_product'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # batch_pricing_row
    op.drop_table('batch_pricing_row')

    # batch_pricing
    op.drop_table('batch_pricing')



# -*- coding: utf-8; -*-
"""add more custorder stuff

Revision ID: 08cc2ef12c18
Revises: 3f3523703313
Create Date: 2021-12-21 17:00:11.841243

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '08cc2ef12c18'
down_revision = '3f3523703313'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # pending_product
    op.add_column('pending_product', sa.Column('vendor_name', sa.String(length=50), nullable=True))
    op.add_column('pending_product', sa.Column('vendor_uuid', sa.String(length=32), nullable=True))
    op.add_column('pending_product', sa.Column('vendor_item_code', sa.String(length=20), nullable=True))
    op.add_column('pending_product', sa.Column('unit_cost', sa.Numeric(precision=10, scale=5), nullable=True))
    op.create_foreign_key('pending_product_fk_vendor', 'pending_product', 'vendor', ['vendor_uuid'], ['uuid'])

    # batch_custorder_row
    op.add_column('batch_custorder_row', sa.Column('pending_product_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key('batch_custorder_row_fk_pending_product', 'batch_custorder_row', 'pending_product', ['pending_product_uuid'], ['uuid'])
    op.add_column('batch_custorder_row', sa.Column('unit_regular_price', sa.Numeric(precision=8, scale=3), nullable=True))
    op.add_column('batch_custorder_row', sa.Column('unit_sale_price', sa.Numeric(precision=8, scale=3), nullable=True))
    op.add_column('batch_custorder_row', sa.Column('sale_ends', sa.DateTime(), nullable=True))
    op.add_column('batch_custorder_row', sa.Column('product_scancode', sa.String(length=14), nullable=True))
    op.add_column('batch_custorder_row', sa.Column('product_item_id', sa.String(length=50), nullable=True))

    # custorder_item
    op.add_column('custorder_item', sa.Column('pending_product_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key('custorder_item_fk_pending_product', 'custorder_item', 'pending_product', ['pending_product_uuid'], ['uuid'])
    op.add_column('custorder_item', sa.Column('unit_regular_price', sa.Numeric(precision=8, scale=3), nullable=True))
    op.add_column('custorder_item', sa.Column('unit_sale_price', sa.Numeric(precision=8, scale=3), nullable=True))
    op.add_column('custorder_item', sa.Column('sale_ends', sa.DateTime(), nullable=True))
    op.add_column('custorder_item', sa.Column('product_scancode', sa.String(length=14), nullable=True))
    op.add_column('custorder_item', sa.Column('product_item_id', sa.String(length=50), nullable=True))


def downgrade():

    # custorder_item
    op.drop_column('custorder_item', 'product_item_id')
    op.drop_column('custorder_item', 'product_scancode')
    op.drop_column('custorder_item', 'sale_ends')
    op.drop_column('custorder_item', 'unit_sale_price')
    op.drop_column('custorder_item', 'unit_regular_price')
    op.drop_constraint('custorder_item_fk_pending_product', 'custorder_item', type_='foreignkey')
    op.drop_column('custorder_item', 'pending_product_uuid')

    # batch_custorder_row
    op.drop_column('batch_custorder_row', 'product_item_id')
    op.drop_column('batch_custorder_row', 'product_scancode')
    op.drop_column('batch_custorder_row', 'sale_ends')
    op.drop_column('batch_custorder_row', 'unit_sale_price')
    op.drop_column('batch_custorder_row', 'unit_regular_price')
    op.drop_constraint('batch_custorder_row_fk_pending_product', 'batch_custorder_row', type_='foreignkey')
    op.drop_column('batch_custorder_row', 'pending_product_uuid')

    # pending_product
    op.drop_constraint('pending_product_fk_vendor', 'pending_product', type_='foreignkey')
    op.drop_column('pending_product', 'unit_cost')
    op.drop_column('pending_product', 'vendor_item_code')
    op.drop_column('pending_product', 'vendor_uuid')
    op.drop_column('pending_product', 'vendor_name')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
File Monitor for Linux
"""

from __future__ import unicode_literals, absolute_import

from six.moves import queue
import logging

import pyinotify

from rattail.daemon import Daemon
from rattail.threads import Thread
from rattail.filemon.config import load_profiles
from rattail.filemon.actions import perform_actions
from rattail.filemon.util import queue_existing


log = logging.getLogger(__name__)


class EventHandler(pyinotify.ProcessEvent):
    """
    Event processor for file monitor daemon.  This receives notifications of
    file system events, and places new files on the queue as appropriate.
    """

    def my_init(self, profile=None, **kwargs):
        self.profile = profile

    def process_IN_ACCESS(self, event):
        log.debug(u"IN_ACCESS: {0}".format(event.pathname))

    def process_IN_ATTRIB(self, event):
        log.debug(u"IN_ATTRIB: {0}".format(event.pathname))

    def process_IN_CLOSE_WRITE(self, event):
        log.debug(u"IN_CLOSE_WRITE: {0}".format(event.pathname))
        if not self.profile.watch_locks:
            self.profile.queue.put(event.pathname)

    def process_IN_CREATE(self, event):
        log.debug(u"IN_CREATE: {0}".format(event.pathname))

    def process_IN_DELETE(self, event):
        log.debug(u"IN_DELETE: {0}".format(event.pathname))
        if self.profile.watch_locks and event.pathname.endswith(u'.lock'):
            self.profile.queue.put(event.pathname[:-5])

    def process_IN_MODIFY(self, event):
        log.debug(u"IN_MODIFY: {0}".format(event.pathname))

    def process_IN_MOVED_TO(self, event):
        log.debug(u"IN_MOVED_TO: {0}".format(event.pathname))
        if not self.profile.watch_locks:
            self.profile.queue.put(event.pathname)


class FileMonitorDaemon(Daemon):
    """
    Linux daemon implementation of the File Monitor.
    """

    def run(self):

        watch_manager = pyinotify.WatchManager()
        notifier = pyinotify.Notifier(watch_manager)

        mask = (pyinotify.IN_ACCESS
                | pyinotify.IN_ATTRIB
                | pyinotify.IN_CLOSE_WRITE
                | pyinotify.IN_CREATE
                | pyinotify.IN_DELETE
                | pyinotify.IN_MODIFY
                | pyinotify.IN_MOVED_TO)

        monitored = load_profiles(self.config)
        for key, profile in monitored.items():

            # Create a file queue for the profile.
            profile.queue = queue.Queue()

            # Perform setup for each of the watched folders.
            for path in profile.dirs:

                # Maybe put all pre-existing files in the queue.
                if profile.process_existing:
                    queue_existing(profile, path)

                # Create a watch for the folder.
                log.debug(u"adding watch to profile '{0}' for folder: {1}".format(key, path))
                watch_manager.add_watch(path, mask, proc_fun=EventHandler(profile=profile))

            # Create an action thread for the profile.
            name = u'actions-{0}'.format(key)
            log.debug(u"starting action thread: {0}".format(name))
            thread = Thread(target=perform_actions, name=name, args=(profile,))
            thread.daemon = True
            thread.start()

        # Fire up the watchers.
        notifier.loop()


def get_daemon(config, pidfile=None):
    """
    Get a :class:`FileMonitorDaemon` instance.
    """
    if pidfile is None:
        pidfile = config.get('rattail.filemon', 'pid_path',
                             default='/var/run/rattail/filemon.pid')
    return FileMonitorDaemon(pidfile, config=config)


def start_daemon(config, pidfile=None, daemonize=False):
    """
    Start the file monitor daemon.
    """
    get_daemon(config, pidfile).start(daemonize)


def stop_daemon(config, pidfile=None):
    """
    Stop the file monitor daemon.
    """
    get_daemon(config, pidfile).stop()



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
New Table Feature
"""

from __future__ import unicode_literals, absolute_import

from rattail.features import Feature
from rattail.features.base import JSONWidget

import colander


class NewTableFeature(Feature):
    """
    New Table Feature
    """
    feature_key = 'new-table'
    feature_title = "New Table"

    def make_schema(self, **kwargs):

        class Column(colander.MappingSchema):

            name = colander.SchemaNode(colander.String())

            data_type = colander.SchemaNode(colander.String())

            nullable = colander.SchemaNode(colander.Boolean())

            description = colander.SchemaNode(colander.String())

        class Columns(colander.SequenceSchema):

            column = Column()

        class Schema(colander.MappingSchema):

            table_name = colander.SchemaNode(colander.String())

            model_name = colander.SchemaNode(colander.String())

            model_title = colander.SchemaNode(colander.String())

            model_title_plural = colander.SchemaNode(colander.String())

            description = colander.SchemaNode(colander.String())

            versioned = colander.SchemaNode(colander.Boolean())

            columns = Columns(widget=JSONWidget())

        return Schema(**kwargs)

    def get_defaults(self):
        return {
            'table_name': 'cool_widget',
            'model_name': 'CoolWidget',
            'model_title': "Cool Widget",
            'model_title_plural': "Cool Widgets",
            'description': "Represents a Cool Widget.",
            'versioned': True,
            'columns': [
                {
                    'name': 'id',
                    'data_type': 'Integer()',
                    'nullable': False,
                    'description': "Numeric ID for the widget.",
                },
                {
                    'name': 'name',
                    'data_type': 'String(length=255)',
                    'nullable': False,
                    'description': "Name of the widget.",
                },
                {
                    'name': 'description',
                    'data_type': 'String(length=255)',
                    'nullable': True,
                    'description': "Description of the widget.",
                },
                {
                    'name': 'coolness_factor',
                    'data_type': 'Numeric(precision=4, scale=3)',
                    'nullable': True,
                    'description': "Just how cool is this widget, from 0.000 to 1.000.",
                },
            ],
        }

    def generate(self, **kwargs):
        context = dict(kwargs)
        context['app_package'] = self.config.app_package()
        context['app_slug'] = context['app_package'] # TODO
        context['envroot'] = '/srv/envs/{}'.format(context['app_slug']) # TODO
        context['prefixed_table_name'] = '{}_{}'.format(
            context['app_prefix'], context['table_name'])
        context['prefixed_model_name'] = '{}{}'.format(
            context['app_cap_prefix'], context['model_name'])

        return self.generate_mako('/new-table/instructions.mako', **context)



# -*- coding: utf-8 -*-
"""grow purchase cost fields

Revision ID: 675a0034becc
Revises: c10adeff4117
Create Date: 2021-10-10 19:57:33.360590

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '675a0034becc'
down_revision = 'c10adeff4117'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_item
    op.alter_column('purchase_item', 'catalog_unit_cost', type_=sa.Numeric(precision=10, scale=5))
    op.alter_column('purchase_item', 'po_unit_cost', type_=sa.Numeric(precision=10, scale=5))
    op.alter_column('purchase_item', 'invoice_case_cost', type_=sa.Numeric(precision=10, scale=5))
    op.alter_column('purchase_item', 'invoice_unit_cost', type_=sa.Numeric(precision=10, scale=5))

    # purchase_credit
    op.alter_column('purchase_credit', 'invoice_case_cost', type_=sa.Numeric(precision=10, scale=5))
    op.alter_column('purchase_credit', 'invoice_unit_cost', type_=sa.Numeric(precision=10, scale=5))

    # batch_purchase_row
    op.alter_column('batch_purchase_row', 'catalog_unit_cost', type_=sa.Numeric(precision=10, scale=5))
    op.alter_column('batch_purchase_row', 'po_unit_cost', type_=sa.Numeric(precision=10, scale=5))
    op.alter_column('batch_purchase_row', 'invoice_case_cost', type_=sa.Numeric(precision=10, scale=5))
    op.alter_column('batch_purchase_row', 'invoice_unit_cost', type_=sa.Numeric(precision=10, scale=5))

    # batch_purchase_credit
    op.alter_column('batch_purchase_credit', 'invoice_case_cost', type_=sa.Numeric(precision=10, scale=5))
    op.alter_column('batch_purchase_credit', 'invoice_unit_cost', type_=sa.Numeric(precision=10, scale=5))


def downgrade():

    # batch_purchase_credit
    op.alter_column('batch_purchase_credit', 'invoice_unit_cost', type_=sa.Numeric(precision=7, scale=3))
    op.alter_column('batch_purchase_credit', 'invoice_case_cost', type_=sa.Numeric(precision=7, scale=3))

    # batch_purchase_row
    op.alter_column('batch_purchase_row', 'invoice_unit_cost', type_=sa.Numeric(precision=7, scale=3))
    op.alter_column('batch_purchase_row', 'invoice_case_cost', type_=sa.Numeric(precision=7, scale=3))
    op.alter_column('batch_purchase_row', 'po_unit_cost', type_=sa.Numeric(precision=7, scale=3))
    op.alter_column('batch_purchase_row', 'catalog_unit_cost', type_=sa.Numeric(precision=7, scale=3))

    # purchase_credit
    op.alter_column('purchase_credit', 'invoice_unit_cost', type_=sa.Numeric(precision=7, scale=3))
    op.alter_column('purchase_credit', 'invoice_case_cost', type_=sa.Numeric(precision=7, scale=3))

    # purchase_item
    op.alter_column('purchase_item', 'invoice_unit_cost', type_=sa.Numeric(precision=7, scale=3))
    op.alter_column('purchase_item', 'invoice_case_cost', type_=sa.Numeric(precision=7, scale=3))
    op.alter_column('purchase_item', 'po_unit_cost', type_=sa.Numeric(precision=7, scale=3))
    op.alter_column('purchase_item', 'catalog_unit_cost', type_=sa.Numeric(precision=7, scale=3))



# -*- coding: utf-8; -*-
"""add table for API Tokens

Revision ID: 4747a017b8f9
Revises: 0aeaac17cd6e
Create Date: 2023-05-14 11:18:12.265992

"""

# revision identifiers, used by Alembic.
revision = '4747a017b8f9'
down_revision = '0aeaac17cd6e'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # user_api_token
    op.create_table('user_api_token',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('user_uuid', sa.String(length=32), nullable=False),
                    sa.Column('description', sa.String(length=255), nullable=False),
                    sa.Column('token_string', sa.String(length=255), nullable=False),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.ForeignKeyConstraint(['user_uuid'], ['user.uuid'], name='user_api_token_fk_user'),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('user_api_token_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('user_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('description', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('token_string', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('created', sa.DateTime(), nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_user_api_token_version_end_transaction_id'), 'user_api_token_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_user_api_token_version_operation_type'), 'user_api_token_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_user_api_token_version_transaction_id'), 'user_api_token_version', ['transaction_id'], unique=False)


def downgrade():

    # user_api_token
    op.drop_index(op.f('ix_user_api_token_version_transaction_id'), table_name='user_api_token_version')
    op.drop_index(op.f('ix_user_api_token_version_operation_type'), table_name='user_api_token_version')
    op.drop_index(op.f('ix_user_api_token_version_end_transaction_id'), table_name='user_api_token_version')
    op.drop_table('user_api_token_version')
    op.drop_table('user_api_token')
